/**
 * @file cfEzgcd.h
 *
 * Extended Zassenhaus GCD over finite fields and Z
**/

#ifndef CF_EZ_GCD_H
#define CF_EZ_GCD_H
CanonicalForm
EZGCD_P (const CanonicalForm& A, const CanonicalForm& B);

CanonicalForm ezgcd ( const CanonicalForm & f, const CanonicalForm & g );
#endif
