/* emacs edit mode for this file is -*- C++ -*- */

/**
 * @file cf_primes.h
 *
 * access to prime tables
**/

#ifndef INCL_CF_PRIMES_H
#define INCL_CF_PRIMES_H

// #include "config.h"

/*BEGINPUBLIC*/

int cf_getPrime( int i );

int cf_getNumPrimes();

int cf_getSmallPrime( int i );

int cf_getNumSmallPrimes();

int cf_getBigPrime( int i );

int cf_getNumBigPrimes();

/*ENDPUBLIC*/

#endif /* ! INCL_CF_PRIMES_H */
