/* emacs edit mode for this file is -*- C++ -*- */

#ifndef INCL_CF_GLOBALS_H
#define INCL_CF_GLOBALS_H

//{{{ docu
//
// cf_globals.h - header to cf_globals.cc.
//
//}}}

// #include "config.h"

#include "cf_switches.h"

/*BEGINPUBLIC*/

extern const char factoryVersion[];
extern const char factoryConfiguration[];

/*ENDPUBLIC*/


#endif /* ! INCL_CF_GLOBALS_H */
